{$G-,I-}

{{$DEFINE DEBUG}

{$IFDEF DEBUG}
{$S+,R+}
{$ELSE}
{$S-,R-}
{$ENDIF}

program Chunk;

uses
  dos,
  support,
  cmdlin,
  objects;

type
  flopTableType=record
    abbr:word;
    realsize:longint;
  end;

const
  maxBufSize:word=65536-16;
  verifyBufSize:word=18*512*2; {tracksize of 1.44MB}
  version='Chunk v0.1 | 20041013 | Quick, simple, fast file splitter | trixter@oldskool.org';
  helpFileName='usage.txt';{name of file to dump help overlay text to}

  maxFlopTableEntries=7;
  flopTable:array[0..maxFlopTableEntries-1] of flopTableType=(
    (abbr:160; realsize:160256),
    (abbr:180; realsize:179712),
    (abbr:320; realsize:322560),
    (abbr:360; realsize:362496),
    (abbr:720; realsize:730112),
    (abbr:1200;realsize:1213952),
    (abbr:1440;realsize:1457664)
  );

  chunkSize:longint=0;     {size of each chunk in bytes}
  numChunks:word=0;        {number of parts to create when splitting}
  Splitting:boolean=true; {if false, we are combining}
  DiskMode:boolean=false; {if true, we are pausing and doing automatic checks}
  AutoSize:boolean=false; {if true, check targ. free space for chunk size}
  Verify:boolean=false;   {if true, read written chunk for readability}
  Compare:boolean=false;  {if true, written chunk is byte-compared to original -- not yet implemented}
  Overwrite:boolean=false;{if true, destination file(s) are overwritten without prompting}
  EraseDisk:boolean=false;{if true, diskette files will be erased before chunks are written}

var
  inStream,outStream:pBufStream;
  inFileName,outFileName:string;

Procedure printHelp;
{procedure to print and/or dump help file attached as overlay to .exe file}

type
  footype=array[0..32767] of char;
  pfootype=^footype;

var
  f:file;
  buf:pfootype;
  w:word;
  helpFileSize:word;
  EXEFileSize:longint;
  dumping:boolean;

begin
  if param_text('h')='w'
    then dumping:=true
    else dumping:=false;
  EXEFileSize:=EXESize(paramstr(0));
  helpFileSize:=SizeOfFile(paramstr(0))-EXEFileSize;
  {are we printing the help file, or just displaying it?}
  getmem(buf,helpFileSize);
  assign(f,paramstr(0));
  reset(f,1);
  helpFileSize:=filesize(f)-EXEFileSize;
  seek(f,EXEFileSize);
  blockread(f,buf^,helpFileSize);
  close(f);
  if dumping then begin
    write('Saving command-line usage to the file "',helpFileName,'"... ');
    assign(f,helpFileName);
    rewrite(f,1);
    blockwrite(f,buf^,helpFileSize);
    close(f);
    writeln('done.');
  end else begin
    for w:=0 to helpFileSize-1 do write(buf^[w]);
  end;
  freemem(buf,helpFileSize);
end;

{$IFDEF DEBUG}
Procedure printdebug;
begin
  writeln('infile: "',inFilename,'"');
  writeln('outfile: "',outFilename,'"');
  writeln('chunksize: ',chunksize,' numchunks: ',numchunks);
  writeln('splitting: ',splitting);
  writeln('diskmode: ',diskmode,' autosize: ',autosize);
  writeln('verify: ',verify,' compare: ',compare,' overwrite: ',overwrite);
  {errorOut('debug printout finished');}
end;
{$ENDIF}

Procedure chunkInit;

var
  l:longint;
  b:byte;
  s:string;

begin
  writeln(version);
  {if asking for help, or both file arguments missing, print help then exit}
  if (non_flag_count<2) or is_param('h') or is_param('?') then begin
    printHelp;
    halt;
  end;
  {if neither splitting nor combining, complain}
  if not is_param('s') and not is_param('f') and not is_param('n') and not is_param('a') {splitting options}
  and not is_param('c') {combining option}
    then errorOut('You must supply either a splitting or combining command-line option.');

  if is_param('d') then DiskMode:=true;

  {determine whether or not we're splitting or combining}
  if is_param('c')
    then Splitting:=false
    else Splitting:=true;

  {based on splitting and/or combining, set the rest of the vars}
  if Splitting then begin {gather splitting vars}
    if is_param('a')
      then if DiskMode
        then AutoSize:=true
        else errorOut('You must enable Disk Mode (/d) to use automatic sizing (/a).');
    if is_param('e')
      then if Diskmode
        then EraseDisk:=true
        else errorOut('You must enable Disk Mode (/d) to enable automatic erasure (/e).');
    if is_param('f') then begin
      l:=param_int('f');
      if l=0
        then AutoSize:=true
        else begin
          chunkSize:=0;
          for b:=0 to maxFlopTableEntries-1 do
            if flopTable[b].abbr=l
              then chunkSize:=flopTable[b].realSize;
        end;
      {nothing?  Then invalid format}
      if (chunkSize=0) and not AutoSize then errorOut('Invalid value for diskette format size!');
    end;
    if is_param('s') then begin
      {does number have K or M?  If so, adjust; if not, use as-is as bytes}
      s:=upstring(param_text('s'));
      if pos('K',s)<>0
        then begin
          dec(s[0]); {chop off the 'K'}
          chunkSize:=StrToInt(s) * 1024;
        end else if pos('M',s)<>0
          then begin
            dec(s[0]); {chop off the 'M'}
            chunkSize:=StrToInt(s) * 1024 * 1024;
          end
          else chunkSize:=param_int('s');
      if (chunkSize=0) and not AutoSize then errorOut('No size given with the /s switch.');
    end;
    if is_param('n') then begin
      numChunks:=param_int('n');
      if (numChunks=0) and not AutoSize then errorOut('Number of chunks not given with the /n switch.');
    end;
    inFileName:=non_flag_param(1); outFileName:=non_flag_param(2);
    if not FileExists(inFileName)
      then errorOut('Source file not found, or has special status or atypical flags set!');
    b:=pos('.',outFileName);
    if b<>0
      then errorOut('No extension allowed in the target when splitting.');
    {splitting by number of files?  Get size of source and calculate chunk size}
    if numChunks<>0
      then begin
        chunkSize:=SizeOfFile(inFileName);
        if chunksize=-1
          then errorOut('Source file suddenly disappeared?')
          else begin
            chunkSize:=chunkSize div numChunks;
            if (SizeOfFile(inFileName) mod numChunks)<>0
              then inc(chunkSize);
          end;
      end;
  end else begin {gather combining vars}
    inFileName:=non_flag_param(1); outFileName:=non_flag_param(2);
    b:=pos('.',inFileName);
    if b<>0
      then errorOut('No extension allowed in the source when combining.');
  end;
  {rest of options}
  if is_param('v') then begin
    Verify:=true;
    if param_text('v')='c'
      then Compare:=true; {not yet implemented}
  end;
  if is_param('o')
    then Overwrite:=true;

  {$IFDEF DEBUG}
  printdebug;
  {$ENDIF}

  if Verify then begin
    s:='Target verification ';
    if compare then s:=s+'with byte comparison ';
    s:=s+'hasn''t been implemented yet.';
    errorOut(s);
  end;
end;

Procedure CheckBufferSize;
const
  margin:longint=128; {margin of ram in bytes to leave free for verification buffer}
var
  l,m:longint; {l is needed to work around BUG IN PASCAL COMPILER, argh!}
begin
  m:=maxavail;
  if Verify then dec(m,verifyBufSize);
  {if running on 128K machine, 64K * 2 ram won't be available, so:}
  l:=maxBufSize;
  l:=(l*2) + margin;
  if m < l then begin
    {print "out of ram" message}
    writeln('Warning:  Low memory, only ',m,' bytes free for copying.  Adjusting...');
    {adjust buffer size to slightly less than half of availble ram}
    maxBufSize:=(m-margin) div 2;
  end;
end;

Function waitForDrive(filename:string;Ready:boolean):boolean;
{returns FALSE if user aborted the wait}
var
  drive:char;
begin
  if pos(':',FileName)=0 then errorOut('Disk mode requires the drive letter to be included in the target name.');
  drive:=FileName[1];
  if ready then begin
    write(#13#10'Waiting for new diskette... (press a key to abort)');
    repeat until DiskReady(drive) or keypressed;
    if keypressed then begin
      repeat readkeychar until not keypressed;
      waitForDrive:=false;
    end else begin
      writeln(#13'New diskette found.                               ');
      waitForDrive:=true;
    end;
  end else begin
    write(#13#10#7'Please remove diskette.');
    repeat until not DiskReady(drive) or keypressed;
    if keypressed then begin
      repeat readkeychar until not keypressed;
      waitForDrive:=false;
    end else begin
      writeln(#13'Diskette removed.      ');
      waitForDrive:=true;
    end;
  end;
end;

Procedure chunkSplit;
{Basic idea is to keep writing chunks of size chunkSize until no more bytes left in source to write}
const
  curChunk:word=0;
  bytesToCopy:longint=0;

var
  outFileNameFull:string;
  DirInfo:SearchRec;
  f:file;

begin
  new(inStream);
  CheckBufferSize;
  inStream^.init(inFileName,stOpenRead,maxBufSize);
  bytesToCopy:=inStream^.getSize;
  while bytesToCopy>0 do begin
    if DiskMode
      then if not waitForDrive(outFileName,true)
        then errorOut('User abort while waiting for new diskette');
    {if /e, erase the contents of the disk}
    if DiskMode and EraseDisk then begin
      writeln(#13'Automatic disk erasure (/e) specified; deleting files on drive ',upcase(outFileName[1]),':');
      FindFirst(outFileName[1]+':\*.*', AnyFile, DirInfo);
      while DosError = 0 do begin
        write('Erasing ',DirInfo.Name,'...');
        assign(f,outFileName[1]+':\'+DirInfo.Name);
        erase(f);
        writeln('done.');
        FindNext(DirInfo);
      end;
    end;
    {check target for free space if AutoSize is requested}
    if AutoSize then begin
      chunkSize:=DiskFree(ord(upcase(outFileName[1]))-64);
      if chunkSize=-1 then errorOut('Target drive invalid?');
      if chunkSize=0 then errorOut('Diskette is full! (use /e to automatically erase all files on target)');
      {$IFDEF DEBUG}
      writeln('found ',chunksize,' bytes free on target');
      {$ENDIF}
    end;
    outFileNameFull:=outFileName+'.'+NumToExt(curChunk);
    if not Overwrite
      then if fileExists(outFileNameFull)
        then errorOut(outFileNameFull+' already exists (use /o to overwrite)');
    new(outStream);
    outStream^.init(outFileNameFull,stCreate,maxBufSize);
    if outStream^.status<>stOK then errorOut('Problem creating '+outFileName);
    if bytesToCopy<chunkSize then chunkSize:=bytesToCopy;
    if inStream^.status<>stOK then errorOut('Input file has invalid status?');
    write(#13'Writing ',chunkSize,' bytes to ',outFileNameFull,'... ');
    outStream^.CopyFrom(inStream^,chunkSize);
    if outStream^.status<>stOK then errorOut('Problem writing to '+outFileNameFull);
    outStream^.done;
    dispose(outStream);
    dec(bytesToCopy,chunkSize);
    inc(curChunk);
    if DiskMode and (bytesToCopy<>0)
      then if not waitForDrive(outFileName,false)
        then errorOut('User abort while waiting for diskette removal');
  end;
  inStream^.done;
  dispose(inStream);
end;

Procedure chunkCombine;
{basic strategy:  Find total number of chunks, then copy them into stream}
var
  curChunk:word;
  inFileNameFull:string;
  lastChunkSize:longint;

begin
  if DiskMode then begin
    writeln('Warning: Cannot determine number of chunks to read in Disk Mode if all chunk');
    writeln('sizes are equal.  If this is the case, you will have to signal the last disk');
    writeln('by pressing a key after the last disk has been read and removed.');
  end;
  if not Overwrite {if we're not overwriting blindly}
    then if fileExists(outFileName) {and file does exist}
      then errorOut(outFileName+' already exists (use /o to overwrite)'); {then abort to prevent overwriting files}
  if DiskMode
    then numChunks:=999
    else while FileExists(inFileName+'.'+NumToExt(numChunks)) do inc(numChunks);
  new(outStream);
  CheckBufferSize;
  outStream^.init(outFileName,stCreate,maxBufSize);
  if outStream^.status<>stOK then errorOut('Problem creating '+outFileName);
  {
  loop in DiskMode will abort if one of three conditions are met:
  1. User indicates last chunk by pressing a key
  2. Chunk is found that is smaller than previous chunks
  3. User loops through all 1000 disks ;-)
  }
  for curChunk:=0 to numChunks-1 do begin
    inFileNameFull:=inFileName+'.'+NumToExt(curChunk);
    if DiskMode
      then if not waitForDrive(inFileNameFull,true)
        then begin
          writeln('User signaled last disk; aborting...');
          break;
        end;
    new(inStream);
    inStream^.init(inFileNameFull,stOpenRead,maxBufSize);
    if inStream^.status<>stOK then errorOut('Input file has invalid status; wrong file?');
    write(#13'Writing ',inStream^.getSize,' bytes from ',inFileNameFull,' to ',outFileName,'... ');
    outStream^.CopyFrom(inStream^,inStream^.getSize);
    if outStream^.status<>stOK then errorOut('Problem writing to '+outFileName);
    if DiskMode then begin
    {did we find last chunk?  If yes, close out; if not, wait for drive}
      if curChunk=0 {if on the first chunk}
        then lastChunkSize:=inStream^.getSize {then get the size}
        else if inStream^.getSize < lastChunkSize then begin {else if cursize < original size}
          writeln(#13#10'Last chunk found (current chunk is smaller than preceeding chunks)...');
          inStream^.done;
          dispose(inStream);
          break;
        end;
      if not waitForDrive(inFileNameFull,false)
        then begin
          writeln('User signaled last disk; aborting...');
          break; {user sez this is the last disk}
        end;
    end;
    {if not DiskMode, then just go about our merry way as usual}
    inStream^.done;
    dispose(inStream);
  end;
  outStream^.done;
  dispose(outStream);
end;

Procedure chunkDone;
begin
  writeln(#13#10'Done.');
end;

begin
  chunkInit;
  if Splitting
    then chunkSplit
    else chunkCombine;
  chunkDone;
end.
